from strictyaml import load, Map, Str, Seq, Optional, MapPattern, Any, YAMLError
from hitchstory import exceptions
from hitchstory.story import Story
from path import Path
import copy


class StoryFile(object):
    """
    YAML file containing one or more named stories, part of a collection.
    """
    def __init__(self, filename, collection):
        self._filename = filename
        self._yaml = filename.bytes().decode('utf8')
        self._collection = collection

        steps_schema = Seq(Str() | MapPattern(Str(), Any(), maximum_keys=1))

        story_schema = {
            Optional("steps"): steps_schema,
            Optional("description"): Str(),
            Optional("based on"): Str(),
            Optional("with"): Any(),
        }

        variation_schema = {
            Optional("steps"): steps_schema,
            Optional("description"): Str(),
            Optional("with"): Any(),
            Optional('given'): self.engine.schema.preconditions,
        }

        if self.engine.schema.about is not None:
            for about_property, property_schema in self.engine.schema.about.items():
                story_schema[about_property] = property_schema

        story_schema[Optional('given')] = self.engine.schema.preconditions
        story_schema[Optional('variations')] = MapPattern(Str(), Map(variation_schema))

        # Load YAML into memory
        try:
            self._parsed_yaml = load(
                self._yaml,
                MapPattern(Str(), Map(story_schema))
            )
            self._updated_yaml = copy.copy(self._parsed_yaml)
        except YAMLError as error:
            raise exceptions.StoryYAMLError(
                filename, str(error)
            )

    @property
    def engine(self):
        return self._collection.engine

    def update(self, story, step, kwargs):
        """
        Update a specific step in a particular story during a test run.
        """
        if story.variation:
            if step.child_index >= 0:
                yaml_story = self._updated_yaml[story.based_on]['variations'][story.child_name]
                if step.arguments.single_argument:
                    yaml_story['steps'][step.child_index][step.name] = \
                        list(kwargs.values())[0]
                else:
                    for key, value in kwargs.items():
                        yaml_story['steps'][step.child_index][step.name][key] = \
                            value
            else:
                yaml_story = self._updated_yaml[story.based_on]
                if step.arguments.single_argument:
                    yaml_story['steps'][step.index][step.name] = \
                        list(kwargs.values())[0]
                else:
                    for key, value in kwargs.items():
                        yaml_story['steps'][step.index][step.name][key] = \
                          value
        else:
            if step.arguments.single_argument:
                self._updated_yaml[story.name]['steps'][step.index][step.name] = \
                    list(kwargs.values())[0]
            else:
                for key, value in kwargs.items():
                    self._updated_yaml[story.name]['steps'][step.index][step.name][key] = \
                      value

    @property
    def filename(self):
        return self._filename

    @property
    def path(self):
        return Path(self._filename)

    def ordered_arbitrarily(self):
        """
        Return all of the stories in the file.
        """
        stories = []
        for name, parsed_main_story in self._parsed_yaml.items():
            stories.append(Story(
                self, str(name), parsed_main_story, self.engine, self._collection
            ))

            variations = self._parsed_yaml[name].get("variations", {}).items()

            for variation_name, parsed_var_name in variations:
                stories.append(
                    Story(
                        self,
                        variation_name,
                        parsed_var_name,
                        self.engine,
                        self._collection,
                        parent=str(name),
                        variation=True,
                    )
                )
        return stories



"""
User-exposed engine related code.
"""
from hitchstory import exceptions
from strictyaml import MapPattern, Str, Any, Map, Validator, Optional
from hitchstory import utils


def validate(**kwargs):
    """
    Decorator for validating arguments in a HitchStory step.
    """
    def decorator(step_function):
        for arg in kwargs:
            if arg not in step_function.__code__.co_varnames:
                raise exceptions.StepContainsInvalidValidator(
                    "Step {} does not contain argument '{}' listed as a validator.".format(
                        step_function.__repr__(), arg
                    )
                )
        step_function._validators = kwargs
        return step_function
    return decorator


def expected_exception(exception_type):
    """
    Tag a method as expecting an exception of exception type.
    """
    def decorator(step_function):
        if not hasattr(step_function, '_expected_exceptions'):
            step_function._expected_exceptions = [exception_type, ]
        else:
            step_function._expected_exceptions.append(exception_type)
        return step_function
    return decorator


class StorySchema(object):
    """
    Represents user-defineable parts of the hitchstory schema:

    * preconditions - properties which set up the story.
    * parameters - variables which you can feed into a story.
    * about - descriptive properties - e.g. feature names, issue ticket numbers
    """
    def __init__(self, given=None, params=None, about=None):
        if given is None:
            self._preconditions = MapPattern(Str(), utils.YAML_Param | Any())
        else:
            _preconditions = {}
            for name, validator in given.items():
                assert isinstance(validator, Validator),\
                    "precondition schema must be strictyaml Validators"
                _preconditions[name] = utils.YAML_Param | validator
            self._preconditions = Map(_preconditions)

        if about is not None:
            assert isinstance(about, dict), "about must be a dict of named validators"
            for key, validator in about.items():
                assert isinstance(key, str) or isinstance(key, Optional), \
                    "name must be a string or strictyaml Optional."
                assert isinstance(validator, Validator), "validator must be strictyaml Validator"

        self._about = about

    @property
    def preconditions(self):
        return self._preconditions

    @property
    def about(self):
        return self._about


class NewStory(object):
    def __init__(self, engine):
        self._engine = engine

    def save(self):
        """
        Write out the updated story to file.
        """
        story_file = self._engine.story.story_file
        story_file.path.write_text(
            story_file._updated_yaml.as_yaml()
        )


class BaseEngine(object):
    schema = StorySchema()

    @property
    def new_story(self):
        return NewStory(self)

    @property
    def given(self):
        return self._preconditions

    def set_up(self):
        pass

    def on_abort(self, signal_num, stack_frame):
        pass

    def on_failure(self, report):
        pass

    def on_success(self):
        pass

    def tear_down(self):
        pass



from strictyaml import Regex
from re import compile
from path import Path
import prettystack


SRC_DIR = Path(__file__).realpath().dirname()

TEMPLATE_DIR = SRC_DIR.joinpath("templates")


DEFAULT_STACK_TRACE = prettystack.PrettyStackTemplate()\
                                 .to_console()\
                                 .cut_calling_code(
                                      SRC_DIR.joinpath("step_method.py")
                                 )


PARAM_REGEX = r"^\(\((.*?)\)\)$"


YAML_Param = Regex(PARAM_REGEX)


def is_parameter(text):
    """
    Is the chunk of YAML data passed to us a parameter?

    i.e. like so (( parametername ))
    """
    return isinstance(text, str) and compile(PARAM_REGEX).match(text) is not None


def parameter_name(text):
    """
    Return parameter name from parameter text.

    e.g. (( param_name )) -> "param_name"
    """
    return compile(PARAM_REGEX).match(text).group(1).strip()


def to_underscore_style(text):
    """Changes "Something like this" to "something_like_this"."""
    text = text.lower().replace(" ", "_").replace("-", "_")
    return ''.join(x for x in text if x.isalpha() or x.isdigit() or x == "_")



from hitchstory.story_list import StoryList
from hitchstory.story_file import StoryFile
from hitchstory.engine import BaseEngine
from hitchstory import exceptions
from slugify import slugify
from path import Path
from copy import copy


class StoryCollection(object):
    """
    Unordered group of related stories.
    """

    def __init__(self, storypaths, engine):
        """
        Create a collection of hitch stories from a list (or generator) of
        story paths and an initialized story engine.
        """
        if not isinstance(engine, BaseEngine):
            raise exceptions.WrongEngineType(
                "Engine should inherit from hitchstory.BaseEngine."
            )
        if isinstance(storypaths, str):
            raise exceptions.InvalidStoryPaths((
                "storypaths should be a list or iterator returning a list of story files"
                " (e.g. using pathquery). Instead it was string '{0}'."
            ).format(storypaths))
        self._storypaths = storypaths
        self._engine = engine
        self._in_filename = None
        self._named = None
        self._filters = []
        self._params = {}
        self._story_files = {}
        self._stories = None
        self._filtered_stories = None

    @property
    def engine(self):
        return self._engine

    def story_file(self, filename):
        if filename not in self._story_files:
            self._story_files[filename] = StoryFile(filename, self)
        return self._story_files[filename]

    @property
    def stories(self):
        if self._stories is None:
            self._stories = {}
            for filename in self._storypaths:
                if not Path(filename).exists():
                    raise exceptions.InvalidStoryPaths(
                        "Story file name '{0}' does not exist.".format(filename)
                    )
                if Path(filename).isdir():
                    raise exceptions.InvalidStoryPaths(
                        "Story path '{0}' is a directory.".format(filename)
                    )
                for story in self.story_file(filename).ordered_arbitrarily():
                    if story.slug in self._stories:
                        raise exceptions.DuplicateStoryNames(story, self._stories[story.slug])
                    self._stories[story.slug] = story
        return self._stories

    def ordered_arbitrarily(self):
        """
        Return a StoryList object containing stories filtered
        from the collection.
        """
        if self._filtered_stories is None:
            self._filtered_stories = []
            all_stories = []
            for story in self.stories.values():
                filtered = True
                for filter_func in self._filters:
                    if not filter_func(story):
                        filtered = False
                if self._named is not None:
                    if story.name != self._named:
                        filtered = False
                if self._in_filename is not None:
                    if Path(story.filename).abspath() != Path(self._in_filename).abspath():
                        filtered = False
                if filtered:
                    self._filtered_stories.append(story)
                all_stories.append(story)

            # Check for non-existent inherited stories
            for story in self._filtered_stories:
                if "based on" in story._parsed_yaml:
                    inherited_from = story._parsed_yaml['based on'].text
                    inherited_from_slug = slugify(inherited_from)
                    found = False
                    for search_story in all_stories:
                        if inherited_from_slug == search_story.slug:
                            found = True
                    if not found:
                        raise exceptions.BasedOnStoryNotFound(
                            inherited_from,
                            story.name,
                            story.filename,
                        )
        return self._filtered_stories

    def filter(self, filter_func):
        assert callable(filter_func)
        new_collection = self.copy()
        new_collection._filters.append(filter_func)
        return new_collection

    def in_filename(self, filename):
        new_collection = self.copy()
        new_collection._in_filename = Path(filename)
        if not new_collection._in_filename.exists():
            raise exceptions.FileNotFound(filename)
        return new_collection

    def with_params(self, **params):
        new_collection = self.copy()
        new_collection._params = params
        return new_collection

    def without_filters(self):
        new_collection = self.copy()
        new_collection._filters = []
        return new_collection

    def in_any_filename(self):
        new_collection = self.copy()
        new_collection._in_filename = None
        return new_collection

    def copy(self):
        new_collection = copy(self)
        new_collection._filtered_stories = None
        new_collection._stories = None

        for story in new_collection._story_files.values():
            story._collection = self
        return new_collection

    def named(self, name):
        """
        Return a single story object with the name specified.

        Only slugified names are compared. E.g. "Story NAME" and "story-name" are equivalent.
        """
        slug = slugify(name)
        for story in self.ordered_arbitrarily():
            if slug == story.slug:
                return story
        raise exceptions.StoryNotFound(name)

    def ordered_by_name(self):
        """
        Return a list of stories ordered by name.
        """
        return StoryList(sorted(self.ordered_arbitrarily(), key=lambda story: story.name))

    def shortcut(self, *words):
        """
        Return a single story that matches all of the words.
        """
        matching = []
        stories = self.ordered_by_name()
        slugified_words = [slugify(word) for word in words]
        for story in stories:
            if len(words) == len([
                slugified_word for slugified_word in slugified_words
                if slugified_word in story.slug
            ]):
                matching.append(story)
        if len(matching) == 0:
            raise exceptions.StoryNotFound(", ".join(words))
        elif len(matching) > 1:
            raise exceptions.MoreThanOneStory(
                "\n".join([
                    "{0} (in {1})".format(story.name, story.filename) for story in matching
                ])
            )
        else:
            return matching[0]

    def one(self):
        stories = self.ordered_by_name()
        if len(stories) > 1:
            raise exceptions.MoreThanOneStory(
                "\n".join([
                    "{0} (in {1})".format(story.name, story.filename) for story in stories
                ])
            )
        elif len(stories) == 0:
            raise exceptions.NoStories()
        else:
            return stories[0]
